package my_context

import (
	"context"
	"errors"
	"log"
	"net/http"
	"net/http/httptest"
	"testing"
	"time"
)

// THIS WAS WHAT WE DID WHEN WE DIDN"T PROPOGATE CONTEXT
// type SpyStore struct {
// 	response string
// 	cancelled bool
// 	t *testing.T
// }

// func (s *SpyStore) assertWasCancelled() {
// 	s.t.Helper()
// 	if !s.cancelled {
// 		s.t.Error("store was not told to cancel")
// 	}
// }

// func (s *SpyStore) assertWasNotCancelled() {
// 	s.t.Helper()
// 	if s.cancelled {
// 		s.t.Error("store was told to cancel")
// 	}
// }

// func (s *SpyStore) Fetch() string {
// 	time.Sleep(100 * time.Millisecond)
// 	return s.response
// }

// func (s *SpyStore) Cancel() {
// 	s.cancelled = true
// }

// func TestServer(t *testing.T) {
// 	data := "Hello, Gopher"
// 	t.Run("Happy path", func(t *testing.T) {
// 		store := &SpyStore{response: data, t: t}
// 		svr := Server(store)

// 		request := httptest.NewRequest(http.MethodGet, "/", nil)
// 		response := httptest.NewRecorder()

// 		svr.ServeHTTP(response, request)

// 		if response.Body.String() != data {
// 			t.Errorf(`got "%s" want "%s"`, response.Body.String(), data)
// 		}

// 		store.assertWasNotCancelled()

// 	})
// 	t.Run("Should cancel", func(t *testing.T) {
// 		store := &SpyStore{response: data, t: t}
// 		svr := Server(store)

// 		request := httptest.NewRequest(http.MethodGet, "/", nil)
// 		// we are copying the parent context but with a new Done channel.
// 		// The returned context's Done channel is closed when the returned cancel
// 		// function is called or when the parent context's Done channel is closed,
// 		// whichever happens first.
// 		cancellingCtx, cancel := context.WithCancel(request.Context())
// 		time.AfterFunc(5*time.Millisecond, cancel)
// 		// changing the current request context to our new cancellingCtx
// 		request = request.WithContext(cancellingCtx)
// 		response := httptest.NewRecorder()

// 		svr.ServeHTTP(response, request)

// 		store.a

// 	})
// }

type SpyStore struct {
	response string
	t        *testing.T
}

func (s *SpyStore) Fetch(ctx context.Context) (string, error) {
	data := make(chan string, 1)

	go func() {
		var result string
		for _, c := range s.response {
			select {
			case <-ctx.Done():
				log.Println("spy store got cancelled")
				return
			default:
				// simulates some network traffic?
				time.Sleep(10 * time.Millisecond)
				result += string(c)
			}
		}
		data <- result
	}()

	select {
	case <-ctx.Done():
		return "", ctx.Err()
	case res := <-data:
		return res, nil
	}
}

// Rolling our own Spy for ResponseWriter because httptest.ResponseRecorder can't be
// used to determine if any response was writing in the error case

// SpyResponse Writer implements http.ResponseWriter so we can use it in the test.
type SpyResponseWriter struct {
	written bool
}

func (s *SpyResponseWriter) Header() http.Header {
	s.written = true
	return nil
}

func (s *SpyResponseWriter) Write([]byte) (int, error) {
	s.written = true
	return 0, errors.New("not implemented")
}

func (s *SpyResponseWriter) WriteHeader(statusCode int) {
	s.written = true
}

func TestServer(t *testing.T) {
	data := "Hello, Gopher"

	t.Run("Happy Path", func(t *testing.T) {

		store := &SpyStore{response: data, t: t}
		svr := Server(store)

		request := httptest.NewRequest(http.MethodGet, "/", nil)
		response := httptest.NewRecorder()

		svr.ServeHTTP(response, request)

		if response.Body.String() != data {
			t.Errorf(`got "%s", want "%s"`, response.Body.String(), data)
		}
	})

	t.Run("tells store to cancel work if request is cancelled", func(t *testing.T) {
		store := &SpyStore{response: data, t: t}
		svr := Server(store)

		request := httptest.NewRequest(http.MethodGet, "/", nil)

		cancellingCtx, cancel := context.WithCancel(request.Context())
		time.AfterFunc(5*time.Millisecond, cancel)
		request = request.WithContext(cancellingCtx)

		response := &SpyResponseWriter{}

		svr.ServeHTTP(response, request)

		if response.written {
			t.Errorf("A response should not have been written")
		}

	})

}
